using Nemerle;
using Nemerle.Imperative;
using Nemerle.IO;
using System;
using System.IO;


namespace Rsdn.ParserTools.Bnf
{
  public module TokenManager
  {
    public Filter(token : Token) : bool
    {
      match (token.TokenType)
      {
        | TokenType.Whitespace 
        | TokenType.Comment
        | TokenType.Error => false
        | _ => true
      }
    }
    
    public Eval(token : Token, _ : IErrorLog) : Node
    {
      def loc = token.Locator;
      def get_string_value(raw)
      {
        def get_part(esc, i, current)
        {
          def get_hex(c)
          {
            if (c >= '0' && c <= '9')
              (c :> int) - ('0' :> int)
            else if (c >= 'A' && c <= 'F')
              (c :> int) - ('A' :> int) + 10
            else
              (c :> int) - ('a' :> int) + 10
          }
          if (i > raw.Length)
            current
          else
          {
            if (esc)
              match (raw[i + 1])
              {
                | 'n'  => get_part(false, i + 2, current + "\n")
                | 't'  => get_part(false, i + 2, current + "\t")
                | 'b'  => get_part(false, i + 2, current + "\b")
                | 'r'  => get_part(false, i + 2, current + "\r")
                | '\\' => get_part(false, i + 2, current + "\\")
                | '\"' => get_part(false, i + 2, current + "\"")
                | 'u' =>
                  def val = get_hex(raw[i + 2]) * 0x1000 + get_hex(raw[i + 3]) * 0x100 +
                            get_hex(raw[i + 4]) * 0x10   + get_hex(raw[i + 5]);
                  get_part(false, i + 6, current + (val :> char).ToString())
                | _    => throw Exception("Internal parser bug???")
              }
            else
              match (raw.IndexOf('\\', i))
              {
                | -1 => current + raw.Substring(i)
                | x  => get_part(true, x, current + raw.Substring(i, x - i))
              }
          }
        }
        get_part(false, 0, string.Empty)
      }
      match (token.TokenType)
      {
        | TokenType.Identifier => 
          if (token.Text[0] != '@')
            Node.Id(loc, token.Text)
          else
            Node.Id(loc, token.Text.Substring(1))
        | TokenType.String     =>
            def rawValue = token.Text.Substring(1, token.Text.Length - 2);
            Node.String(loc, get_string_value(rawValue))
        | _                    => Node.Token(loc)
      }
    }
  }
}