using Nemerle.Extensions;
using Nemerle.Utility;
using NUnit.Framework;
using System;

// ToDo:
// - promote test_ranges local functions to private method
// - tests for Insert
// - tests for RemoveRange
namespace Rsdn.ParserTools.Tests.RleList
{
  [TestFixture]
  public class RleListTests
  {
    //[Test]
    //public AddNulls() : void
    //{
      //def empty = CreateEmptyList();
      //for (mutable i = 0; i < 3; i++)
      //{
        //empty.Add(null);
      //}
      //def arr = empty.ToArray();
      //Assert.AreEqual(array[0, 0, 0], arr);
    //}
    
    [Test]
    public GetAllRanges() : void
    {
      def test_ranges(list, expectedCount)
      {
        Assert.AreEqual(expectedCount, list.GetRanges().ToArray().Length);
      }
      
      test_ranges(CreateEmptyList(), 0);
      test_ranges(CreateList(5), 1);
      test_ranges(CreateListWithRepeatedValues(5), 5);
    }
    
    [Test]
    public GetRanges() : void
    {
      def test(list, range, expectedCount)
      {
        Assert.AreEqual(expectedCount, list.GetRanges(range).ToArray().Length);
      }
      
      def list = CreateListWithRepeatedValues(4);
      test(list, ListRange(0, 1), 1);
      test(list, ListRange(0, 2), 2);
      test(list, ListRange(0, 3), 2);
      test(list, ListRange(0, 4), 3);
      test(list, ListRange(3, 6), 1);
      test(list, ListRange(3, 7), 2);
      test(list, ListRange(6, list.Count), 1);
    }
    
    #region FillTests
    
    [Test]
    public Fill() : void
    {
      def test_ranges(list, expectedLength)
      {
        def ranges = list.GetRanges().ToArray();
        Assert.AreEqual(expectedLength, ranges.Length);
      }
      
      def list = CreateListWithRepeatedValues(3);

      list.Fill(1, 2, 1);
      test_ranges(list, 2);
      Assert.AreEqual(array[1, 1, 1, 3, 3, 3], list.ToArray());
      
      list.Fill(3, 3, 1);
      test_ranges(list, 1);
      Assert.AreEqual(array[1, 1, 1, 1, 1, 1], list.ToArray());
      
      list.Fill(0, 2, 2);
      test_ranges(list, 2);
      Assert.AreEqual(array[2, 2, 1, 1, 1, 1], list.ToArray());
      
      list.Fill(4, 2, 2);
      test_ranges(list, 3);
      Assert.AreEqual(array[2, 2, 1, 1, 2, 2], list.ToArray());
    }
    
    [Test]
    public FillWithSameValues() : void
    {
      def test(list, range, value, expectedValues, expectedRange)
      {
        list.Fill(range, value);
        def ranges = list.RangesOf(value).ToArray();
        
        Assert.AreEqual(1, ranges.Length);
        Assert.AreEqual(expectedRange, ranges[0]);
        Assert.AreEqual(expectedValues, list.ToArray());
      }
      
      def list = CreateListWithRepeatedValues(4);
      def expectedValues = array[1, 2, 2, 3, 3, 3, 4, 4, 4, 4];

      test(list, ListRange(0, 1), 1, expectedValues, ListRange(0, 1));
      test(list, ListRange(3, 6), 3, expectedValues, ListRange(3, 6));
      test(list, ListRange(6, 9), 4, expectedValues, ListRange(6, 10));
    }
    
    //[Test]
    //public FillWithNulls() : void
    //{
      //def filled = CreateList.[object](10);
    //}
    
    [Test]
    public FillInvalidRange() : void
    {
      def try_fill(list, range)
      {
        try
        {
          list.Fill(range, 0);
          Assert.Fail("Must throw exception");
        }
        catch
        {
          | e is ArgumentException => ()
          | _ => Assert.Fail("Must throw System.ArgumentException")
        }
      }
      
      def empty = CreateEmptyList();
      def filled = CreateList(10);
      
      try_fill(empty, ListRange(0, 1));
      try_fill(filled, ListRange(-2, -1));
      try_fill(filled, ListRange(-1, 0));
      try_fill(filled, ListRange(0, 11));
      try_fill(filled, ListRange(12, 13));
    }
    
    #endregion // FillTests
    
    private static CreateEmptyList[a]() : RleList[a]
    {
      RleList();
    }
    
    private static CreateList[a](n : int) : RleList[a]
    {
      def list = RleList();
      for (mutable i = 0; i < n; i++)
      {
        list.Add(DefaultValue(a));
      }
      list
    }
    
    private static CreateListWithRepeatedValues(n : int) : RleList[int]
    {
      def list = RleList();
      for (mutable i = 1; i <= n; i++)
      {
        repeat(i)
        {
          list.Add(i);
        }
      }
      list
    }
  }
}
