﻿using Nemerle.Collections;
using System;
using System.Text;
using Rsdn.ParserTools.Bnf;
using Rsdn.ParserTools.Finite;


namespace Rsdn.ParserTools
{
  [Record]
  public class KeywordInfo
  {
    public TextList : list[string];
    public TargetId : string;
  }

  [Record]
  public class DfaInfo
  {
    public Lexer : Node.Section;
    public Keywords : IConstListDictionary[string, KeywordInfo];
    public Dfa : Dfa[DfaState];
    
    public static CreateDfaInfo(lexer : Node.Section, log : IErrorLog) : DfaInfo
    {
      def detect_keywords(regexps)
      {
        def isTrivial(id)
        {
          def isUnion(node)
          {
            | Node.Union(a, b) => isUnion(a) && isUnion(b)
            | Node.String(_) => true
            | _ => false
          }
          match (lexer.body[id].expr)
          {
            | Node.String(_) => true
            | Node.Union(_, _) as u => isUnion(u)              
            | _ => false
          }
        }
        def get_keyword_list(expr)
        {
          | Node.Union(a, b) => List.Append(get_keyword_list(a), get_keyword_list(b))
          | Node.String(s) => [s]
          | _ => throw Exception()
        }
        def nonTrivialRegexps = ListSet(lexer.body.Keys.Filter(k => !isTrivial(k))).ToMap(k => regexps[k]);
        def nfa = NfaBuilder.FromRegexps(nonTrivialRegexps);
        def dfa = DfaBuilder.FromNfa(nfa);
        def activeDfa = ActiveDfa(dfa, nonTrivialRegexps.Keys);
        def keywords = ListDictionary();
        foreach (id when isTrivial(id) in lexer.body.Keys)
        {
          def keywordList = get_keyword_list(lexer.body[id].expr);
          def match_keywords(l, n)
          {
            match (l)
            {
              | []   => n
              | h::t =>
                def name = ActiveDfa.Match(activeDfa, h);
                if (n == null || n == name)
                  match_keywords(t, name)
                else
                  null
            }
          }
          def name = match_keywords(keywordList, null);
          when (name != null)
            keywords.Add(id, KeywordInfo(keywordList, name));
        }
        keywords
      }
      
      def regexps = BnfUtils.BuildRegexps(lexer, log);
      def keywords = detect_keywords(regexps);
      def nonKeywordRegexps = regexps.Filter((k, _) => !keywords.Keys.Contains(k));
      def nfa = NfaBuilder.FromRegexps(nonKeywordRegexps);
      def dfa = DfaBuilder.FromNfa(nfa);
      def minDfa = DfaBuilder.Minimize(dfa);
      DfaInfo(lexer, keywords, minDfa)
    }
  }
}