using Nemerle;
using Nemerle.Collections;
using System.IO;


namespace Rsdn.ParserTools
{
  public enum ErrorLevel
  {
    | Info
    | Hint
    | Warning
    | Error
  }
  
  [Record]
  public class Error
  {
    public Locator : TextLocator;
    public ErrorLevel : ErrorLevel;
    public Message : string;
  }
  
  public interface IErrorLog
  {
    Message(error : Error) : void;
    SetPanic() : void;
    Panic : bool { get; }
    HasErrors : bool { get; }
  }
  
  public class SimpleErrorLog : IErrorLog
  {
    errors : Vector[Error] = Vector.[Error]();
    mutable panic = false;
    mutable hasErrors = false;
    
    public Message(error : Error) : void
    {
      errors.Add(error);
      when (error.ErrorLevel == ErrorLevel.Error) hasErrors = true;
    }
    
    public SetPanic() : void
    {
      panic = true;
    }
    
    public Panic : bool
    {
      get
      {
        panic
      }
    }
    
    public HasErrors : bool
    {
      get
      {
        hasErrors
      }
    }
    
    public Write(writer : TextWriter) : void
    {
      foreach (error in errors)
      {
        writer.WriteLine($"$(error.ErrorLevel) at $(error.Locator): $(error.Message)");
      }
    }
  }
}