using Nemerle.Collections;
using Nemerle.Extensions;
using System;
using System.Collections.Generic;
using System.Text;


namespace Rsdn.ParserTools.Finite
{
  public class ActiveDfa
  {
    dfa : Dfa[DfaState];
    transitions : IConstListDictionary[DfaState,  array[int * int * DfaState]];
    stateNames : IConstListDictionary[DfaState, string];
    mutable current : DfaState;
    
    public this(dfa : Dfa[DfaState], regexps : IConstListSet[string])
    {
      this.dfa = dfa;
      def transitions = ListDictionary();
      def lists = ListMultiDictionary();
      foreach ((src, range, dest) in dfa.Transitions)
        foreach (subrange in range.IncludedRanges)
          lists.Add(src, (subrange.First, subrange.Last - 1, dest));
      def stateNames = ListDictionary();
      foreach (state in lists.Keys)
      {
        def vector = Vector(lists[state]);
        vector.Sort((a, b) => a[0].CompareTo(b[0]));
        transitions.Add(state, vector.ToArray());
        def ids = Vector(state.RegexpIDs);
        ids.Sort((a, b) => regexps.IndexOf(b).CompareTo(regexps.IndexOf(a)));
        if (ids.Count > 0)
          stateNames.Add(state, ids[0])
        else
          stateNames.Add(state, null)
      }
      this.transitions = transitions;
      this.stateNames = stateNames;
      Reset();
    }
    
    public Reset() : void
    {
      current = dfa.Initial;
    }
    
    public CurrentName : string
    {
      get
      {
        if (current == null)
          null
        else
          stateNames[current]
      }
    }
    
    public Current : DfaState
    {
      get
      {
        current
      }
    }
    
    public Next(ch : int) : void
    {
      def findNext(l, u)
      {
        if (l > u)
          null
        else
        {
          def i = (l + u) / 2;
          def (from, to, state) = transitions[current][i];
          if (ch < from)
            findNext(l, i - 1)
          else if (ch > to)
            findNext(i + 1, u)
          else
            state
        }
      }
      unless (current == null)
        current = findNext(0, transitions[current].Length - 1);
    }
    
    public static Match(dfa : ActiveDfa, str : string) : string
    {
      def loop(str)
      {
        if (str == string.Empty)
          dfa.CurrentName
        else if (dfa.Current == null)
          null
        else
        {
          dfa.Next(str[0] :> int);
          loop(str.Substring(1))
        }
      }
      dfa.Reset();
      loop(str)
    }
  }
}