using System;
using System.Collections.Generic;
using System.Text;

namespace Rsdn.ParserTools.Finite
{
  public class TransitionFunction['s]
  {
    private transitions : ListDictionary['s, ListMultiDictionary['s, CharRange]];
    
    public this()
    {
      transitions = ListDictionary();
    }
    
    public this(transitions : IEnumerable['s * CharRange * 's])
    {
      this();
      Append(transitions);
    }
    
    public Transitions : IEnumerable[('s * CharRange * 's)]
    {
      get
      {
        foreach (initial in transitions.Keys)
          foreach (final in transitions[initial].Keys)
            foreach (condition in transitions[initial][final])
            {
              yield (initial, condition, final);
            }
      }
    }
    
    public Add(initial : 's, condition : CharRange, final : 's) : void
    {
      when (!transitions.ContainsKey(initial))
      {
        transitions.Add(initial, ListMultiDictionary());
      }
      transitions[initial].Add(final, condition);
    }
    
    public Append(transitions : IEnumerable['s * CharRange * 's]) : void
    {
      foreach (transition in transitions)
      {
        this.Add(transition);
      }
    }
    
    public Append(other : TransitionFunction['s]) : void
    {
      Append(other.Transitions);
    }
    
    public GetTransitions(state : 's, condition : CharRange) : IConstListSet['s]
    {
      def targets = ListSet();
      foreach ((cond, target) in GetTransitions(state))
      {
        when ((condition == CharRange.Empty && cond == CharRange.Empty) ||
              (condition != CharRange.Empty && cond.Contains(condition)))
        {
          targets.Add(target);
        }
      }
      targets
    }
    
    public GetConditions(state : 's) : IConstListSet[CharRange]
    {
      def conditions = ListSet();
      foreach ((cond, _) in GetTransitions(state))
      {
        conditions.Add(cond);
      }
      conditions
    }
    
    private GetTransitions(state : 's) : IEnumerable[CharRange * 's]
    {
      when (transitions.ContainsKey(state))
      {
        def transition = transitions[state];
        when (transition != null)
        {
          foreach (target in transition.Keys)
            foreach (condition in transition[target])
            {
              yield (condition, target);
            }
        }
      }
    }
  }
}