using Nemerle;
using Nemerle.Imperative;
using Nemerle.Utility;
using System;
using SCG = System.Collections.Generic;


namespace Rsdn.ParserTools.Grammar
{
  public class LrItem
  {
    public Production : Production;
    public Index : int;
    public Current : list[Symbol * string * TextLocator];
    hash = 0;
    
    public this(prod : Production)
    {
      this.Production = prod;
      this.Index = 0;
      this.Current = prod.Sentence;
      hash = prod.GetHashCode();
    }
    
    protected this(proto : LrItem)
    {
      this.Production = proto.Production;
      this.Index = proto.Index + 1;
      this.Current = match(proto.Current)
      {
        | [] => throw ArgumentException()
        | _ :: y => y
      }
      hash = (proto.GetHashCode() << 1) ^ 23561236;
    }
    
    public GetNext() : LrItem
    {
      LrItem(this)
    }
    
    public HasNext : bool
    {
      get
      {
        match (this.Current)
        {
          | [] => false
          | _ :: _ => true
        }
      }
    }
    
    public CurrentSymbol : Symbol
    {
      get
      {
        match (this.Current)
        {
          | []     => null
          | x :: _ => x[0]
        }
      }
    }
    
    [OverrideObjectEquals]
    public Equals(other : LrItem) : bool
    {
      this == other
    }
    
    public override GetHashCode() : int
    {
      this.Production.GetHashCode() ^ (this.Index << 4)
    }
    
    public static @==(a : LrItem, b : LrItem) : bool
    {
      object.ReferenceEquals(a, b) || (a.Production == b.Production && a.Index == b.Index)
    }
    
    public static @!=(a : LrItem, b : LrItem) : bool
    {
      !(a == b)
    }
  }
  
  public class LrItemSet
  {
    items : ConstListSet[LrItem];
    
    public this(items : IConstListSet[LrItem])
    {
      this.items = ConstListSet(items);
    }
    
    public Items : IConstListSet[LrItem]
    {
      get
      {
        items
      }
    }
    
    public static @==(a : LrItemSet, b : LrItemSet) : bool
    {
      when (object.ReferenceEquals(a, b)) return true;
      a.items == b.items
    }
    
    public static @!=(a : LrItemSet, b : LrItemSet) : bool
    {
      !(a == b)
    }
    
    [OverrideObjectEquals]
    public Equals(other : LrItemSet) : bool
    {
      this == other
    }
    
    public override GetHashCode() : int
    {
      items.GetHashCode()
    }
  }
  
  public class SlrSystem
  {
    public States : IConstListSet[LrItemSet];
    public Transitions : IConstListMultiDictionary[LrItemSet, Symbol * LrItemSet];
    
    public this(
      states : IConstListSet[LrItemSet],
      transitions : IConstListMultiDictionary[LrItemSet, Symbol * LrItemSet])
    {
      this.States = ConstListSet(states);
      def t = ListMultiDictionary();
      foreach (key in transitions.Keys)
        t.AddMany(key, transitions[key]);
      this.Transitions = t
    }
  }
  
  public class LalrSystem : SlrSystem
  {
    public Lookahead : array[ListMultiDictionary[LrItem, Terminal]];
    
    public this(
      states : IConstListSet[LrItemSet],
      transitions : IConstListMultiDictionary[LrItemSet, Symbol * LrItemSet],
      lookahead : SCG.IEnumerable[ListMultiDictionary[LrItem, Terminal]])
    {
      base(states, transitions);
      this.Lookahead = lookahead.ToArray();
    }
  }
}