using Nemerle;
using Nemerle.Imperative;


namespace Rsdn.ParserTools.Grammar
{
  public abstract class Symbol
  {
    mutable needHash = true;
    mutable hashCode = 0;
        
    public abstract Name : string { get; }
    
    public override Equals(obj : object) : bool
    {
      | other when object.ReferenceEquals(this, other) => true
      | other is Symbol => this.Name == other.Name
      | _  => false
    }
    
    public override ToString() : string
    {
      Name
    }
    
    public override GetHashCode() : int
    {
      when (needHash)
      {
        hashCode = Name.GetHashCode();
        needHash = false;
      }
      hashCode
    }
  }
  
  public abstract class Terminal : Symbol
  {
  }
  
  public abstract class Nonterminal : Symbol
  {
  }
  
  public class ConcreteTerminal : Terminal
  {
    name : string;
    
    public this(name : string)
    {
      this.name = name;
    }
    
    public override Name : string
    {
      get
      {
        name
      }
    }
  }
  
  public class ConcreteNonterminal : Nonterminal
  {
    name : string;
    
    public this(name : string)
    {
      this.name = name;
    }
    
    public override Name : string
    {
      get
      {
        name
      }
    }
  }
  
  public class EmptyTerminal :  Terminal
  {
    static mutable instance : Terminal;
    
    protected this()
    {
    }
    
    public static Instance : Terminal
    {
      get
      {
        when (instance == null)
        {
          instance = EmptyTerminal();
        }
        instance
      }
    }
    
    public override Name : string
    {
      get
      {
        "'e";
      }
    }
  }
  
  public class EofTerminal : Terminal
  {
    static mutable instance : Terminal;
    
    protected this()
    {
    }
    
    public static Instance : Terminal
    {
      get
      {
        when (instance == null)
        {
          instance = EofTerminal()
        }
        instance
      }
    }
    
    public override Name : string
    {
      get
      {
        "Eof"
      }
    }
  }
  
  public class StartNonterminal : Nonterminal
  {
    static mutable instance : Nonterminal;
    
    protected this()
    {
    }
    
    public static Instance : Nonterminal
    {
      get
      {
        when (instance == null)
        {
          instance = StartNonterminal()
        }
        instance
      }
    }
    
    public override Name : string
    {
      get
      {
        "S_"
      }
    }
  }
}