using Nemerle;
using Nemerle.Imperative;
using SCG = System.Collections.Generic;


namespace Rsdn.ParserTools
{
  public interface IConstListSet['a] : SCG.IEnumerable['a]
  {
    Item[index : int] : 'a { get; }
    IndexOf(item : 'a) : int;
    Count : int { get; }
  }
  
  public interface IConstListDictionary['a, 'b] : SCG.IEnumerable['a * 'b]
  {
    Item[key : 'a] : 'b { get; }
    Keys : IConstListSet['a] { get; }
    Count : int { get; }
  }
  
  public interface IConstListMultiDictionary['a, 'b] : SCG.IEnumerable['b]
  {
    Item[key : 'a] : IConstListSet['b] { get; }
    Keys : IConstListSet['a] { get; }
    Count : int { get; }
  }
  
  public module ListSetUtils
  {
    public GetOriginal['a](this source : IConstListSet['a], key : 'a) : 'a
    {
      source[source.IndexOf(key)]
    }
    
    public Contains['a](this source : IConstListSet['a], item : 'a) : bool
    {
      source.IndexOf(item) != -1
    }
    
    public IsSubsetOf['a](this source : IConstListSet['a], set : IConstListSet['a]) : bool
    {
      if (source.Count > set.Count)
        false
      else
      {
        foreach (item in source)
          unless (set.Contains(item)) return false;
        true
      }
    }
    
    public Filter['a](this source : IConstListSet['a], predicate : 'a -> bool) : SCG.IEnumerable['a]
    {
      foreach (item when predicate(item) in source)
        yield item;
    }
    
    public Filter['a, 'b](
      this source : IConstListDictionary['a, 'b],
      predicate : 'a * 'b -> bool) : IConstListDictionary['a, 'b]
    {
      def result = ListDictionary();
      foreach (k when predicate(k, source[k]) in source.Keys)
        result.Add(k, source[k]);
       result;
    }
    
    public Map['a, 'b, 'c, 'd](
      this source : IConstListDictionary['a, 'b],
      f : 'a * 'b -> 'c * 'd) : IConstListDictionary['c, 'd]
    {
      def result = ListDictionary();
      foreach (k in source.Keys)
        result.Add(f(k, source[k]));
      result;
    }
    
    public ToMap['a, 'b](this source : IConstListSet['a], f : 'a -> 'b) : IConstListDictionary['a, 'b]
    {
      def result = ListDictionary();
      foreach (k in source)
        result.Add(k, f(k));
      result
    }
    
    public Union['a](a : IConstListSet['a], b : IConstListSet['a]) : SCG.IEnumerable['a]
    {
      foreach (e in a)
        yield e;
      foreach (e when !a.Contains(e) in b)
        yield e;
    }
  }
}