using Nemerle;
using Nemerle.Extensions;
using System;
using SCG = System.Collections.Generic;

namespace Rsdn.ParserTools
{
  public class RleArray['a] : SCG.IEnumerable['a], ICloneable
  {
    list : RleList['a] = RleList.['a]();
    firstIndex : int;
    
    protected this(firstIndex : int, list : RleList['a] )
    {
      this.firstIndex = firstIndex;
      this.list = list;
    }

    public this(firstIndex : int, lastIndex : int)
    {
      this.firstIndex = firstIndex;
      list.Insert(0, lastIndex - firstIndex + 1, DefaultValue('a));
    }

    public this(range : ListRange)
    {
      this.firstIndex = range.First;
      list.Insert(0, range.Count, DefaultValue('a));
    }
    
    public Clone() : object
    {
      RleArray(firstIndex, list.Clone() :> RleList['a])
    }
    
    public Smooth(size : int) : void
    {
      list.Smooth(size);
    }
    
    public RangesAt(from : int, count : int) : SCG.IEnumerable[ListRange]
    {
      foreach (range in list.RangesAt(from - firstIndex, count))
        yield ListRange.FromCount(range.First + firstIndex, range.Count);
    }
    
    public RangesAt(range : ListRange) : SCG.IEnumerable[ListRange]
    {
      RangesAt(range.First, range.Count)
    }

    public Count : int
    {
      get
      {
        list.Count
      }
    }

    public FirstIndex : int
    {
      get
      {
        firstIndex
      }
    }

    public LastIndex : int
    {
      get
      {
        firstIndex + list.Count - 1
      }
    }

    public All : ListRange
    {
      get
      {
        ListRange.FromCount(firstIndex, list.Count)
      }
    }
    
    public AllRanges : SCG.IEnumerable[ListRange]
    {
      get
      {
        foreach (range in list.AllRanges)
        {
          yield ListRange.FromCount(range.First + firstIndex, range.Count)
        }
      }
    }

    public Item[index : int] : 'a
    {
      get
      {
        list[index - firstIndex]
      }
      set
      {
        list[index - firstIndex] = value
      }
    }

    public Fill(value : 'a) : void
    {
      Fill(value, All);
    }

    public Fill(value : 'a, range : ListRange) : void
    {
      Fill(value, range.First, range.Count);
    }

    public Fill(value : 'a, from : int, count : int) : void
    {
      list.Fill(from - firstIndex, count, value);
    }

    public Convert(range : ListRange, converter : 'a -> 'a) : void
    {
      list.Convert(
        ListRange.FromCount(range.First - firstIndex, range.Count),
        converter);
     }

    public Convert(converter : 'a -> 'a) : void
    {
      Convert(All, converter)
    }
    
    public GetDifferentValues() : SCG.IEnumerable['a]
    {
      list.GetDifferentValues();
    }
    
    public RangesOf(value : 'a) : SCG.IEnumerable[ListRange]
    {
      foreach (range in list.RangesOf(value))
      {
        yield ListRange.FromCount(range.First + firstIndex, range.Count)
      }
    }
    
    public SetOf(value : 'a) : RleArray[bool]
    {
      RleArray(firstIndex, list.SetOf(value));
    }
    
    public IndicesOf(value : 'a) : SCG.IEnumerable[int]
    {
      foreach (range in RangesOf(value))
      {
        for (mutable i = range.First; i < range.Last; i++)
        {
          yield i;
        }
      }
    }
    
    public IsPlain() : bool
    {
      IsPlain(All);
    }
    
    public IsPlain(range : ListRange) : bool
    {
      IsPlain(range.First, range.Count);
    }
    
    public IsPlain(index : int, count : int) : bool
    {
      list.IsPlain(index - firstIndex, count)
    }
    
    [OverrideObjectEquals]
    public Equals(other : RleArray['a]) : bool
    {
      (other.firstIndex == this.firstIndex) && this.list.Equals(other.list)
    }
    
    public override GetHashCode() : int
    {
      list.GetHashCode() ^ firstIndex
    }

    #region IEnumerable['a] Members

    public GetEnumerator() : SCG.IEnumerator['a]
    {
      list.GetEnumerator();
    }

    #endregion
    
    public RangesCount : int
    {
      get
      {
        list.RangesCount
      }
    }
  }
}