using Nemerle;
using Nemerle.Collections;
using Nemerle.Extensions;
using Nemerle.Imperative;
using Nemerle.Utility;
using System;
using SCG = System.Collections.Generic;


namespace Rsdn.ParserTools
{
  public class RleList['a] : SCG.IList['a], ICloneable
  {
    entries : Vector.[int * 'a] = Vector.[int * 'a]();
    
    public this()
    {
      entries.Add((0, DefaultValue('a)))
    }
    
    public Clone() : object
    {
      def clone = RleList();
      clone.entries.Clear();
      foreach (e in entries)
        clone.entries.Add(e);
      clone
    }
    
    EntryIndex(index : int) : int
    {
      def searchRange(l, u)
      {
        if (l > u)
        {
          throw ArgumentException("Invalid index specified.")
        }
        else
        {
          def i = (l + u) / 2;
          def (start, _) = entries[i];
          def (end, _) = entries[i + 1];
          match (index)
          {
            | t when t >= end  => searchRange(i + 1, u)
            | t when t < start => searchRange(l, i - 1)
            | _                => i
          }
        }
      }
      
      match (index)
      {
        | i when i < 0      => -1
        | i when i >= Count => entries.Count
        | _                 => searchRange(0, entries.Count - 1)
      }
    }
    
    public Fill(range : ListRange, item : 'a) : void
    {
      Fill(range.First, range.Count, item)
    }
    
    public Fill(index : int, count : int, item : 'a) : void
    {
      def contains(entry, item)
      {
        if (entry < 0 || entry >= entries.Count - 1)
        {
          false
        }
        else
        {
          def (_, entryItem) = entries[entry];
          entryItem.Equals(item)
        }
      }
      
      def replace(index, count, newItems) : void
      {
        if (newItems.Count > count)
        {
          entries.InsertRange(
            index + count,
            Algorithms.Repeat((0, DefaultValue('a)), newItems.Count - count));
        }
        else
        {
          entries.RemoveRange(index + newItems.Count, count - newItems.Count);
        }
        Algorithms.Copy(newItems, entries, index);
      }
      
      def to = index + count;
      if (index < 0 || to > Count)
      {
        throw ArgumentException("Invalid range specified");
      }
      else
      {
        def firstEntry = EntryIndex(index - 1);
        def lastEntry = EntryIndex(to);
        def (normFrom, normFirst) = if (contains(firstEntry, item))
        {
          def (newFrom, _) = entries[firstEntry];
          (newFrom, firstEntry)
        }
        else
        {
          (index, EntryIndex(index))
        }
        def (normTo, normLast) = if (contains(lastEntry, item))
        {
          def (newTo, _) = entries[lastEntry + 1];
          (newTo, lastEntry)
        }
        else
        {
          (to, EntryIndex(to - 1))
        }
        when (normFirst <= normLast)
        {
          def span = Vector.[int * 'a]();
          def (firstIndex, firstItem) = entries[normFirst];
          when (normFrom > firstIndex) span.Add((firstIndex, firstItem));
          span.Add((normFrom, item));
          def (lastIndex, _) = entries[normLast + 1];
          def (_, lastItem) = entries[normLast];
          when (normTo < lastIndex) span.Add((normTo, lastItem));
          replace(normFirst, normLast - normFirst + 1, span);
        }
      }
    }
    
    public Insert(range: ListRange, item: 'a) : void
    {
      Insert(range.First, range.Count, item)
    }
    
    public Insert(index : int, count : int, item : 'a) : void
    {
      when (count > 0)
      {
        when (index < 0 || index > Count)
        {
          throw ArgumentException("Invalid range specified")
        }
        if (entries.Count > 1)
        {
          def entry = EntryIndex(index - 1);
          for (mutable i = entry + 1; i < entries.Count; ++i)
          {
            def (entry, entryItem) = entries[i];
            entries[i] = (
              if (i == 0) entry else entry + count,
              entryItem);
          }
        }
        else
        {
          entries.Add((count, DefaultValue('a)));
        }
        Fill(index, count, item)
      }
    }
    
    public RemoveRange(range : ListRange) : void
    {
      RemoveRange(range.First, range.Count);
    }
    
    public RemoveRange(index : int, count : int) : void
    {
      def to = index + count;
      if (to > Count || index < 0)
      {
        throw ArgumentException("Invalid range specified");
      }
      else
      {
        def first = EntryIndex(index);
        def last = EntryIndex(to);
        def garbageCount = last - first - 1;
        when (garbageCount > 0)
        {
          entries.RemoveRange(first + 1, garbageCount);
        }
        for (mutable i = first + 1; i < entries.Count; ++i)
        {
          def (spanIndex, spanItem) = entries[i];
          entries[i] = (spanIndex - count, spanItem);
        }
      }
    }
    
    public IsPlain() : bool
    {
      entries.Count < 3;
    }
    
    public IsPlain(range : ListRange) : bool
    {
      IsPlain(range.First, range.Count);
    }
    
    public IsPlain(index : int, count : int) : bool
    {
      EntryIndex(index) == EntryIndex(index + count - 1);
    }
    
    public All : ListRange
    {
      get
      {
        ListRange(0, Count);
      }
    }
    
    public AllRanges : SCG.IEnumerable[ListRange]
    {
      get
      {
        for (mutable i = 0; i < entries.Count - 1; i++)
        {
          yield ListRange(entries[i][0], entries[i + 1][0]);
        }
      }
    }
    
    public GetRanges() : SCG.IEnumerable[ListRange]
    {
      GetRanges(All)
    }
    
    public GetRanges(range : ListRange) : SCG.IEnumerable[ListRange]
    {
      GetRanges(range.First, range.Count)
    }
    
    public GetRanges(index : int, count : int) : SCG.IEnumerable[ListRange]
    {
      def first = EntryIndex(index);
      def last = EntryIndex(index + count - 1);
      for (mutable i = first; i <= last; ++i)
      {
        def (from, _) = entries[i];
        def (to, _) = entries[i + 1];
        yield ListRange(from, to);
      }
    }
    
    public RangesOf(item : 'a) : SCG.IEnumerable[ListRange]
    {
      for (mutable i = 0; i < entries.Count - 1; ++i)
      {
        def (from, testItem) = entries[i];
        when (item.Equals(testItem))
        {
          def (to, _) = entries[i + 1];
          yield ListRange(from, to);
        }
      }
    }
    
    public RangesAt(from : int, count : int) : SCG.IEnumerable[ListRange]
    {
      def to = from + count;
      def fromIndex = EntryIndex(from);
      def toIndex = EntryIndex(to - 1);
      for (mutable i = fromIndex; i <= toIndex; ++i)
      {
        def (a, _) = entries[i];
        def (b, _) = entries[i + 1];
        yield ListRange(
          if (a > from) a else from,
          if (b < to) b else to);
      }
    }
    
    public RangesAt(range : ListRange) : SCG.IEnumerable[ListRange]
    {
      RangesAt(range.First, range.Count)
    }
    
    public SetOf(item : 'a) : RleList[bool]
    {
      def result = RleList.[bool]();
      mutable lastIndex = 0;
      foreach (range in RangesOf(item))
      {
        when (range.First == 0) result.entries.Clear();
        result.entries.Add((range.First, true));
        result.entries.Add((range.Last, false));
        lastIndex = range.Last;
      }
      when (lastIndex < Count)
      {
        result.entries.Add((Count, false));
      }
      result
    }
    
    public GetPlainRange(index : int) : ListRange
    {
      def entry = EntryIndex(index);
      ListRange(entries[entry][0], entries[entry + 1][0])
    }
    
    public Convert(range : ListRange, converter : 'a -> 'a) : void
    {
      def convert(range)
      {
        when (range.Count > 0)
        {
          def subrange = GetPlainRange(range.First);
          def last = if (subrange.Last <= range.Last) subrange.Last else range.Last;
          def current = ListRange(range.First, last);
          
          if (current == subrange)
          {
            def entry = EntryIndex(current.First);
            entries[entry] = (current.First, converter(entries[entry][1]));
          }
          else
          {
            Fill(current, converter(this[current.First]));
          }
          
          convert(ListRange(current.Last, range.Last))
        }
      }
      convert(range)
    }
    
    public Convert(converter : 'a -> 'a) : void
    {
      Convert(All, converter)
    }
    
    public GetDifferentValues() : SCG.IEnumerable['a]
    {
      def listSet = ListSet();
      for (mutable i = 0; i < entries.Count - 1; ++i)
      {
        def (_, item) = entries[i];
        listSet.Add(item);
      }
      listSet
    }
    
    [OverrideObjectEquals]
    public Equals(other : RleList['a]) : bool
    {
      when (object.ReferenceEquals(this, other))
        return true;
      if (entries.Count != other.entries.Count || Count != other.Count)
        false
      else
      {
        for (mutable i = 0; i < entries.Count - 1; ++i)
        {
          def (index, item) = entries[i];
          def (index', item') = other.entries[i];
          when (index != index' || !item.Equals(item'))
            return false;
        }
        true
      }
    }
    
    public override GetHashCode() : int
    {
      mutable hash = 0;
      for (mutable i = 0; i < entries.Count - 1; ++i)
      {
        def (index, item) = entries[i];
        hash <<= 1;
        hash ^= index ^ item.GetHashCode();
      }
      hash
    }
    
    public IndicesOf(item : 'a) : SCG.IEnumerable[int]
    {
      foreach (range in RangesOf(item))
      {
        for (mutable i = range.First; i < range.Last; ++i)
        {
          yield i;
        }
      }
    }
    
    public Smooth(size : int) : void
    {
      unless (entries.Count < 2)
      {
        def newEntries = Vector();
        mutable lastIndex = 0;
        for (mutable i = 1; i < entries.Count; ++i)
        {
          def (b, _) = entries[i];
          def (a, val) = entries[i - 1];
          when (b - a >= size)
          {
            newEntries.Add(lastIndex, val);
            lastIndex = b;
          }
        }
        when (newEntries.Count == 0) newEntries.Add(entries[0]);
        newEntries.Add(entries[entries.Count - 1]);
        entries.Clear();
        entries.AddRange(newEntries);
      }
    }
    
    #region IList['a] interface implementation
    
    public Item[index : int] : 'a
    {
      get
      {
        def (_, result) = entries[EntryIndex(index)];
        result
      }
      set
      {
        Fill(index, 1, value);
      }
    }
    
    public IndexOf(item : 'a) : int
    {
      foreach (test in IndicesOf(item))
      {
        return test;
      }
      -1
    }
    
    public Insert(index : int, item : 'a) : void
    {
      Insert(index, 1, item)
    }
    
    public RemoveAt(index: int) : void
    {
      RemoveRange(index, 1)
    }
    
    #endregion
    
    #region ICollection['a] interface implementation    
    
    public Count : int
    {
      get
      {
        def (result, _) = entries[entries.Count - 1];
        result
      }
    }
    
    public IsReadOnly : bool
    {
      get
      {
        false
      }
    }
    
    public Add(item : 'a) : void
    {
      Insert(Count, 1, item)
    }
    
    public Clear() : void
    {
      entries.Clear();
      entries.Add((0, DefaultValue('a)))
    }
    
    public Contains(item : 'a) : bool
    {
      IndexOf(item) != -1
    }
    
    public CopyTo(dest : array['a], index : int) : void
    {
      if (index + Count > dest.Length)
      {
        throw ArgumentException()
      }
      else
      {
        mutable i = index;
        foreach (item in this)
        {
          dest[i] = item;
          ++i
        }
      }
    }
    
    public Remove(item : 'a) : bool
    {
      def index = IndexOf(item);
      if (index == -1)
      {
        false
      }
      else
      {
        RemoveRange(index, 1);
        true
      }
    }
    
    #endregion    
    
    #region System.Collections.Generic.IEnumerable['a] interface implementation
    
    public GetEnumerator() : SCG.IEnumerator['a]
    {
      for (mutable i = 0; i < entries.Count - 1; ++i)
      {
        def (last, _) = entries[i + 1];
        def (first, item) = entries[i];
        for (mutable j = first; j < last; ++j)
          yield item;
      }
    }
    
    #endregion
    
    public RangesCount : int
    {
      get
      {
        entries.Count - 1;
      }
    }
  }
}