using Nemerle;


namespace Rsdn.ParserTools
{
  [Record]
  public struct TextLocation
  {
    public Row : int;
    public Column : int;
    public static Empty : TextLocation = TextLocation(-1, -1);
    
    is_empty() : bool
    {
      Row < 0 || Column < 0
    }
    
    [OverrideObjectEquals]
    public Equals(other : TextLocation) : bool
    {      
      this == other
    }
    
    public static @==(a : TextLocation, b : TextLocation) : bool
    {
      if (a.is_empty() && b.is_empty())
        true
      else
        (a.Row, a.Column).Equals((b.Row, b.Column))
    }
    
    public static @!=(a : TextLocation, b : TextLocation) : bool
    {
      !(a == b)
    }
    
    public override ToString() : string
    {
      $"($(Row + 1), $(Column + 1))"
    }
    
    // TODO: GetHashCode()
  }
  
  [Record]
  public struct TextRange
  {
    public First : TextLocation;
    public Last : TextLocation;
    public static Empty : TextRange =
      TextRange(TextLocation.Empty, TextLocation.Empty);
    
    public this(firstRow : int, firstColumn : int, lastRow : int, lastColumn : int)
    {
      this.First = TextLocation(firstRow, firstColumn);
      this.Last = TextLocation(lastRow, lastColumn);
    }
    
    [OverrideObjectEquals]
    public Equals(other : TextRange) : bool
    {
      this == other
    }
    
    public static @==(a : TextRange, b : TextRange) : bool
    {
      (a.First, a.Last).Equals((b.First, b.Last))
    }
    
    public static @!=(a : TextRange, b : TextRange) : bool
    {
      !(a == b)
    }
    
    public override ToString() : string
    {
      $"$First - $Last"
    }
    
    // TODO: GetHashCode()
  }
  
  [Record]
  public struct TextLocator
  {
    public static Empty : TextLocator = TextLocator(null, TextRange.Empty);
    public Url : string;
    public Range : TextRange;
    
    public this(url : string, first : TextLocation, second : TextLocation)
    {
      this(url, TextRange(first, second))
    }
    
    [OverrideObjectEquals]
    public Equals(other : TextLocator) : bool
    {
      this == other
    }
    
    public static @==(a : TextLocator, b : TextLocator) : bool
    {
      (a.Url, a.Range).Equals((b.Url, b.Range))
    }
    
    public static @!=(a : TextLocator, b : TextLocator) : bool
    {
      !(a == b)
    }
    
    public First : TextLocation
    {
      get
      {
        Range.First
      }
    }
    
    public Last : TextLocation
    {
      get
      {
        Range.Last
      }
    }
    
    public IsEmpty : bool
    {
      get
      {
        Url == null && Range == TextRange.Empty
      }
    }
    
    public override ToString() : string
    {
      if (IsEmpty)
        string.Empty
      else if (Url == null)
        $"$Range"
      else if (Range == TextRange.Empty)
        $"$Url"
      else
        $"$Url, $Range"
    }
  }
}