<?xml version="1.0" encoding="UTF-8" ?>
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
  <xsl:output method="text" omit-xml-declaration="yes" encoding="utf-8"/>
  
  <xsl:template match="root">
using Nemerle.Collections;
using Nemerle.Imperative;
using System;
using System.IO;
using System.IO.Compression;
using System.Text;
using Rsdn.ParserTools;
<xsl:apply-templates select="using" />
namespace <xsl:value-of select="@namespace" />
{
<xsl:apply-templates select="*"/>
}</xsl:template>
  

  <xsl:template match="root/fa">
  public class <xsl:value-of select="//@tokenizerhelper"/>
  {
    static stateNames : array[array[<xsl:value-of select="//@tokentype"/>]] = array[
<xsl:apply-templates select="names-group" mode="names"/>
    ];
    static automata : array[array[array[int * int * int]]];
    static initialState : int * int * <xsl:value-of select="//@tokentype"/> = (0, 0, <xsl:value-of select="//@tokentype"/>.Eof);
    reader : TextReader;<xsl:if test="@cachetext">
    strBuilder : StringBuilder = StringBuilder();</xsl:if>
    url : string;
    mutable lastToken : <xsl:value-of select="//@tokentype"/>;
    mutable mode : int;
    mutable ch : int;
    mutable row : int;
    mutable column : int;
    mutable currentState : int;
    mutable cr : bool;
        
    static this()
    {
      def stream = typeof(<xsl:value-of select="//@tokenizerhelper"/>).Assembly.GetManifestResourceStream(typeof(<xsl:value-of select="//@tokenizerhelper"/>), "<xsl:value-of select="@resourcefile"/>");
      def decompressor = DeflateStream(stream, CompressionMode.Decompress);
      def reader = StreamReader(decompressor);
      def automataCount = Int32.Parse(reader.ReadLine());
      automata = array(automataCount);
      for (mutable index = 0; index &lt; automataCount; ++index)
      {
        def stateCount = Int32.Parse(reader.ReadLine());
        def states = array(stateCount);
        for (mutable n = 0; n &lt; stateCount; ++n)
        {
          def transitionsCount = Int32.Parse(reader.ReadLine());
          def transitions = array(transitionsCount);
          for (mutable i = 0; i &lt; transitionsCount; ++i)
          {
            def from = Int32.Parse(reader.ReadLine());
            def to = Int32.Parse(reader.ReadLine());
            def state = Int32.Parse(reader.ReadLine());
            transitions[i] = (from, to, state);
          }
          states[n] = transitions;
        }
        automata[index] = states;
      }
      reader.Close();
    }
        
    public this(reader : TextReader, url : string)
    {
       this.reader = reader;
       this.url = url;
       ch = 0;
       column = -1;
       get_next_char();
    }
        
    public static InitialState : object
    {
      get
      {
        return initialState;
      }
    }
        
    get_next_char() : void
    {
      if (ch == 13)
      {
        cr = true;
        ++row;
        column = 0;
      }
      else if (ch == 10)
      {
        column = 0;
        when (!cr) ++row;
        cr = false;
      }
      else
      {
        ++column;
        cr = false;
      }
      ch = reader.Read();
    }
        
    public Mode : int
    {
      get
      {
        return this.mode;
      }
      set
      {
        this.mode = value;
      }
    }
        
    public LastTokenType : <xsl:value-of select="//@tokentype"/>
    {
      get
      {
        return this.lastToken;
      }
    }
        
    public IsEof : bool
    {
      get
      {
        return ch == -1;
      }
    }
        
    public Save() : object
    {
      (currentState, mode, lastToken);
    }
        
    public Reset(state : object, row : int, column : int) : void
    {
      def extState = state :> int * int * <xsl:value-of select="//@tokentype"/>;
      this.currentState = extState[0];
      this.lastToken = extState[2];
      this.mode = extState[1];
      this.row = row;
      this.column = column;
    }
    
    TransformToken(token : TokenType, text : string) : TokenType
    {
      match (mode)
      {
        <xsl:apply-templates select="names-group" mode="transform" />
        | _ => token
      }
    }
    
    public GetToken() : <xsl:value-of select="//@token"/>
    {
      def get_next_state(mode, state, c)
      {
        def s = automata[mode][state];
        def search(l, u)
        {
          if (l > u)
            -1
          else
          {
            def i = (l + u) / 2;
            if (c &lt; s[i][0])
              search(l, i - 1)
            else if (c &gt; s[i][1])
              search(i + 1, u)
            else
              s[i][2]
          }
        }
        search(0, s.Length - 1)
      }
      if (ch == -1)
      {
         <xsl:value-of select="//@token"/>(
         <xsl:value-of select="//@tokentype"/>.Eof,
          TextLocator(url, TextRange(row, column, row, column))<xsl:if test="@cachetext">,
          string.Empty</xsl:if>)
      }
      else
      {
        def firstRow = row;
        def firstColumn = column;
        mutable lastRow = row;
        mutable lastColumn = column;<xsl:if test="@cachetext">
        _ = strBuilder.Remove(0, strBuilder.Length);</xsl:if>
        mutable lastState = currentState;
        while (true)
        {
          lastRow = row;
          lastColumn = column;
          def nextState = get_next_state(mode, currentState, ch);
          when (nextState == -1) break;
          currentState = nextState;
          when (ch == -1) break;
          lastState = nextState;<xsl:if test="@cachetext">
          _ = strBuilder.Append(ch :&gt; char);</xsl:if>
          get_next_char();
        }
        when (stateNames[mode][currentState] == <xsl:value-of select="//@tokentype"/>.Error)
        {
          get_next_char();
        }
        lastToken = stateNames[mode][currentState];
        if (ch == -1)
        {
          currentState = lastState;
        }
        else
        {
          currentState = 0;
        }
        def text = strBuilder.ToString();
        <xsl:value-of select="//@token"/>(
          TransformToken(lastToken, text),
          TextLocator(url, TextRange(firstRow, firstColumn, lastRow, lastColumn))<xsl:if test="@cachetext">,
          text</xsl:if>);
      }
    }
  }
  <xsl:if test="//lr and @generateFacade">
  public class <xsl:value-of select="@facade" /> : ITokenSource
  {
    helper : <xsl:value-of select="//@tokenizerhelper" />;
    
    public this(reader : TextReader, url : string)
    {
      this.helper = <xsl:value-of select="//@tokenizerhelper" />(reader, url);
    }
    
    public GetNextToken() : <xsl:value-of select="//@token" />
    {
      def loop(token)
      {
        match (token.TokenType)
        {
          | TokenType.Eof => token
          | _ when <xsl:value-of select="//@tokenmanager" />.Filter(token) => token
          | _ => loop(helper.GetToken())
        }
      }
      loop(helper.GetToken())
    }
    
    public EvalToken(token : Token, log : IErrorLog) : Node
    {
      <xsl:value-of select="//@tokenmanager" />.Eval(token, log)
    }
  }</xsl:if>
  
  [Record]
  public class <xsl:value-of select="//@token"/>
  {
    public TokenType : <xsl:value-of select="//@tokentype"/>;
    public Locator : TextLocator;<xsl:if test="@cachetext">
    public Text : string;</xsl:if>

    public this(proto : <xsl:value-of select="//@token"/>)
    {
      this.TokenType = proto.TokenType;<xsl:if test="@cachetext">
      this.Text = proto.Text;</xsl:if>
      this.Locator = proto.Locator;
    }
        
    public this(proto : <xsl:value-of select="//@token"/>, tokenType : <xsl:value-of select="//@tokentype"/>)
    {
      this.TokenType = tokenType;<xsl:if test="@cachetext">
      this.Text = proto.Text;</xsl:if>
      this.Locator = proto.Locator;
    }
  }
  
  public enum <xsl:value-of select="//@tokentype"/>
  {
<xsl:apply-templates select="enum"/>
    | Eof
    | Error
  }
</xsl:template>
  
  
  <xsl:template match="item"><xsl:text xml:space="preserve">    | </xsl:text><xsl:value-of select="text()"/><xsl:text xml:space="preserve">
</xsl:text>
  </xsl:template>
  
  <xsl:template match="root/lr">
  public interface ITokenSource
  {
    GetNextToken() : <xsl:value-of select="//@token" />;
    EvalToken(token : <xsl:value-of select="//@token" />, log : IErrorLog) : <xsl:value-of select="//@node" />;
  }
  
  public interface INodeBuilder
  {
    <xsl:apply-templates select="constructor" />
  }

  public class <xsl:value-of select="//@parser "/>
  {
    private enum Action : byte
    {
      | Error
      | Shift
      | Reduce
      | Accept
    }

    static actionTable : array[array[int * Action * int]];
    static gotoTable : array[array[int * int]];
    mutable panic = false;

    static ProductionInfo(prod : int) : int * int
    {<xsl:apply-templates select="grammar"/>
      | _ => throw Exception("Internal parser error")
    }

    static TokenToInt(token : <xsl:value-of select="//@tokentype" />) : int
    {<xsl:apply-templates select="tokentypes/*" />
      | _ => throw Exception("Internal parser error")
    }

    static ReduceProduction(loc : TextLocator, builder : INodeBuilder, prod : int, stack : array[<xsl:value-of select="//@node" />]) : <xsl:value-of select="//@node" />
    {
      match (prod)
      {<xsl:apply-templates select="reductions/*" />
        | _ => throw Exception("Internal parser error")
      }
    }

    static ToBool(node : <xsl:value-of select="//@node" />) : bool
    {
      | <xsl:value-of select="//@node" />.<xsl:value-of select="//@node-null" />() => false
      | _ => true
    }

    static this()
    {
      def stream = typeof(<xsl:value-of select="//@parser"/>).Assembly.GetManifestResourceStream(typeof(<xsl:value-of select="//@parser"/>), "<xsl:value-of select="@resource-name"/>");
      def decompressor = DeflateStream(stream, CompressionMode.Decompress);
      def reader = StreamReader(decompressor);
      def stateCount = Int32.Parse(reader.ReadLine());
      actionTable = array(stateCount);
      gotoTable = array(stateCount);
      for (mutable i = 0; i &lt; stateCount; ++i)
      {
        def cellCount = Int32.Parse(reader.ReadLine());
        def row = array(cellCount);
        for (mutable j = 0; j &lt; cellCount; ++j)
        {
          def symbol = Int32.Parse(reader.ReadLine());
          def action = Int32.Parse(reader.ReadLine()) :> Action;
          def param = Int32.Parse(reader.ReadLine());
          row[j] = (symbol, action, param);
        }
        actionTable[i] = row;
        def gotoCellCount = Int32.Parse(reader.ReadLine());
        def gotoRow = array(gotoCellCount);
        for (mutable j = 0; j &lt; gotoCellCount; ++j)
        {
          def symbol = Int32.Parse(reader.ReadLine());
          def param = Int32.Parse(reader.ReadLine());
          gotoRow[j] = (symbol, param);
        }
        gotoTable[i] = gotoRow;
      }
    }

    public Panic : bool
    {
      get
      {
        return panic;
      }
    }

    public ResetPanic() : void
    {
      panic = false;
    }        

    public Parse(source : ITokenSource, builder : INodeBuilder, log : IErrorLog) : <xsl:value-of select="//@node"/>
    {
      mutable token;
      def stack = Vector.[int * int * <xsl:value-of select="//@node"/> * <xsl:value-of select="//@token"/> * <xsl:value-of select="//@token"/>]();
      def next_token()
      {
        token = source.GetNextToken();
      }
      def state()
      {
        if (stack.Count > 0)
          stack[stack.Count - 1][0]
        else
          0
      }
      def search_array['a, 'b](e : 'a, arr : array['b], comparer)
      {
        def search(l, u)
        {
          def i = (l + u) / 2;
          if (l > u)
            -1
          else
          {
            match (comparer(e, arr[i]))
            {
              | -1  => search(l, i - 1)
              | 1   => search(i + 1, u)
              | _   => i
            }
          }
        }
        search(0, arr.Length - 1)
      }
      def get_action(state, symbol)
      {
        def row = actionTable[state];
        def col = search_array(symbol, row, fun(a, b) { a.CompareTo(b[0]) });
        if (col >= 0) (row[col][1], row[col][2]) else (Action.Error, -1)
      }
      def get_goto(state, symbol)
      {
        def row = gotoTable[state];
        def col = search_array(symbol, row, fun(a, b) { a.CompareTo(b[0]) });
        row[col][1]
      }
      def shift(state)
      {
        def symbol = TokenToInt(token.TokenType);
        def data = source.EvalToken(token, log);
        stack.Add((state, symbol, data, token, token));
        next_token()
      }
      def reduce(prodIndex)
      {
        def (nonterm, len) = ProductionInfo(prodIndex);
        def get_data(i, acc)
        {
          if (i &lt; 0)
            acc
          else
          {
            def (_, _, data, _, _) = stack[stack.Count - len + i];
            get_data(i - 1, data :: acc)
          }
        }
        def data = get_data(len - 1, []);
        def (firstToken, lastToken) =
          if (len > 0)
            (stack[stack.Count - len][3], stack[stack.Count - 1][4])
          else
            (token, token);
        def loc = TextLocator(firstToken.Locator.Url, TextRange(firstToken.Locator.First, lastToken.Locator.Last));
        def newData = ReduceProduction(loc, builder, prodIndex, data.ToArray());
        stack.RemoveRange(stack.Count - len, len);
        def newState = get_goto(state(), nonterm);
        stack.Add((newState, nonterm, newData, firstToken, lastToken));
      }
      def error()
      {
        def search_state(i)
        {
          if (i &lt; -1)
            i
          else
          {
            def testState = if (i > 0) stack[i][0] else 0;
            def (action, _) = get_action(testState, TokenToInt(token.TokenType));
            match (action)
            {
              | Action.Error => search_state(i - 1)
              | _            => i
            }
          }
        }
        panic = true;
        mutable firstToken = token;
        while (true)
        {
          def stackIndex = search_state(stack.Count - 1);
          if (stackIndex >= -1)
          {
            firstToken = if (stackIndex + 1 &lt; stack.Count)
              stack[stackIndex + 1][3]
            else
              token;
            stack.RemoveRange(stackIndex + 1, stack.Count - stackIndex - 1);
            break;
          }
          else
            next_token();
          when (token.TokenType == TokenType.Eof) break;
        }
        def loc = TextLocator(firstToken.Locator.Url, firstToken.Locator.First, token.Locator.Last);
        log.Message(Error(loc, ErrorLevel.Error, "Invalid syntax"));
        token.TokenType != TokenType.Eof
      }
      def loop()
      {
        def (action, param) = get_action(state(), TokenToInt(token.TokenType));
        match (action)
        {
          | Action.Shift  => shift(param); loop()
          | Action.Reduce => reduce(param); loop()
          | Action.Accept => stack[0][2]
          | Action.Error  => if (error()) loop() else stack[0][2]
        }
      }
      next_token();
      loop()
    }
  }
</xsl:template>
  
  <xsl:template match="names" mode="names">      array[
<xsl:apply-templates select="name"/>
      ]<xsl:if test="position()&lt;count(../*)">,
</xsl:if>
</xsl:template>

  <xsl:template match="name">
    <xsl:text xml:space="preserve">        </xsl:text>
    <xsl:value-of select="//@tokentype"/>.<xsl:value-of select="text()"/>
    <xsl:if test="position()&lt;count(../*)">,
</xsl:if>
  </xsl:template>
  
  <xsl:template match="names" mode="transform">
        | <xsl:value-of select="@num" /> =>
          match (token)
          {
            <xsl:apply-templates select="terminal" />
            | _ => token
          }
  </xsl:template>
  
  <xsl:template match="terminal">
            | <xsl:value-of select="//@tokentype" />.<xsl:value-of select="@id" /> =>
              match (text)
              {
                <xsl:apply-templates select="keyword" />
                | _ => token
              }
  </xsl:template>
  
  <xsl:template match="keyword">
                | "<xsl:value-of select="@text" />" => <xsl:value-of select="//@tokentype" />.<xsl:value-of select="@id" />
  </xsl:template>  
  
  <xsl:template match="constructor">
    <xsl:value-of select="@id" />(loc : TextLocator<xsl:apply-templates select="field" />) : <xsl:value-of select="//@node" />;
  </xsl:template>
  
  <xsl:template match="constructor/field">
    <xsl:choose>
      <xsl:when test="@type='node'">, <xsl:value-of select="@id" /> : <xsl:value-of select="//@node" /></xsl:when>
      <xsl:when test="@type='bool'">, <xsl:value-of select="@id" /> : bool</xsl:when>
    </xsl:choose>
  </xsl:template>

  <xsl:template match="using">using <xsl:value-of select="text()"/>;
</xsl:template>

  <xsl:template match="production">
      | <xsl:value-of select="@index" /> => (<xsl:value-of select="@symbol" />, <xsl:value-of select="@length" />)</xsl:template>

  <xsl:template match="tokentype">
      | <xsl:value-of select="//@tokentype" />.<xsl:value-of select="@name" /> => <xsl:value-of select="@value" /></xsl:template>
  
  <xsl:template match="reduction">
        // reduction: <xsl:value-of select="@rule" />
        | <xsl:value-of select="@index" /> =><xsl:text>
          </xsl:text>
    <xsl:choose>
      <xsl:when test="@isfunction">
        builder.<xsl:value-of select="@option" />(loc, <xsl:apply-templates select="arg" />)
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="//@node" />.<xsl:value-of select="@option" />(loc, <xsl:apply-templates select="arg" />)
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template match="copy">
        // copy: <xsl:value-of select="@rule" />
        | <xsl:value-of select="@index" /> =>
          stack[<xsl:value-of select="@copyindex" />]
  </xsl:template>
  
  <xsl:template match="e-reduction">
        // e-reduction: <xsl:value-of select="@rule" />
        | <xsl:value-of select="@index" /> =>
          <xsl:value-of select="//@node" />.<xsl:value-of select="//@node-null" />(loc)
  </xsl:template>
  
  <xsl:template match="arg">
    <xsl:choose>
      <xsl:when test="@type='bool'">
        <xsl:choose>
          <xsl:when test="@index=-1">false</xsl:when>
          <xsl:otherwise>ToBool(stack[<xsl:value-of select="@index" />])</xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      <xsl:when test="@type='node'">
        <xsl:choose>
          <xsl:when test="@index=-1"><xsl:value-of select="//@node" />.<xsl:value-of select="//@node-null" />(loc)</xsl:when>
          <xsl:otherwise>stack[<xsl:value-of select="@index" />]</xsl:otherwise>
        </xsl:choose>
      </xsl:when>
    </xsl:choose>
    <xsl:if test="position()&lt;count(../*)">, </xsl:if>
  </xsl:template>
  
  <xsl:template match="*" mode="index">
  </xsl:template>
</xsl:stylesheet>