﻿using Nemerle;
using Nemerle.Macros;
using Nemerle.Collections;
using Nemerle.Compiler;
using Nemerle.Compiler.Parsetree;
using Nemerle.Utility;
using System;
using System.Text;
using Rsdn.ParserTools.Bnf;
using Rsdn.ParserTools.Grammar;


namespace Rsdn.ParserTools
{
  [MacroUsage(MacroPhase.BeforeInheritance, MacroTargets.Assembly)]
  public macro Bnf(expr)
  {
    match (expr)
    {
      | PExpr.MacroCall(_, _, SyntaxElement.RawToken(rootToken) :: []) =>
        def tokenSource = MacroTokenSource(rootToken);
        def log = MacroErrorLog();
        def parser = Rsdn.ParserTools.Bnf.Parser();
        def rootNode = parser.Parse(tokenSource, log);
        def env = ImplicitCTX().Env;
        def unit = BnfUtils.ApplySubst(rootNode :> Node.Root, log);
        def dfas = ListDictionary();
        unless (env.Manager.IsIntelliSenseMode)
          foreach (lexerId in unit.lexers.Keys)
            dfas.Add(lexerId, DfaInfo.CreateDfaInfo(unit.lexers[lexerId], log));
        BnfMacroHelper.DefineTokenClass(env);
        BnfMacroHelper.DefineTokenTypeClass(env, unit);
        BnfMacroHelper.DefineLexerClass(env, unit, dfas, log);
      | _ => Message.Error(expr.loc, "Expected 'bnf { <definitions> }'")
    }
  }
}