﻿using Nemerle.Collections;
using Nemerle.Compiler;
using System;
using System.Text;
using BNF = Rsdn.ParserTools.Bnf;


namespace Rsdn.ParserTools
{
  class MacroTokenSource : BNF.ITokenSource
  {
    tokenList : Vector.[BNF.Token];
    mutable index = 1;
    
    public this(root : Token)
    {
      tokenList = Vector();
      foreach (token in BnfMacroHelper.GetTokens(root))
      {
        def loc = BnfMacroHelper.ConvertLocation(token.Location);
        def (tokenType, text) = 
          match (token)
          {
            | Token.CharLiteral(ch) => (BNF.TokenType.String, ch.ToString())
            | Token.Comma() => (BNF.TokenType.Comma, ",")
            | Token.Comment(text) => (BNF.TokenType.Comment, text)
            | Token.BeginBrace(_) => (BNF.TokenType.LeftBrace, "{")
            | Token.BeginRound() => (BNF.TokenType.LeftParen, "(")
            | Token.BeginSquare() => (BNF.TokenType.LeftBracket, "[")
            | Token.EndBrace(_) => (BNF.TokenType.RightBrace, "}")
            | Token.EndRound() => (BNF.TokenType.RightParen, ")")
            | Token.EndSquare() => (BNF.TokenType.RightBracket, "]")
            | Token.Identifier("subst") => (BNF.TokenType.Subst, "subst")
            | Token.Identifier("lexer") => (BNF.TokenType.Lexer, "lexer")
            | Token.Identifier(id) => (BNF.TokenType.Identifier, id)
            | Token.Keyword(text) => (BNF.TokenType.Identifier, text)
            | Token.Operator("|") => (BNF.TokenType.VBar, "|")
            | Token.Operator("*") => (BNF.TokenType.Star, "*")
            | Token.Operator("+") => (BNF.TokenType.Plus, "+")
            | Token.Operator("::=") => (BNF.TokenType.Assign, "::=")
            | Token.Operator("?") => (BNF.TokenType.Question, "?")
            | Token.Operator("..") => (BNF.TokenType.Ellipsis, "..")
            | Token.Operator("&") => (BNF.TokenType.Amp, "&")
            | Token.Operator("~") => (BNF.TokenType.Tilde, "~")
            | Token.Operator("\\") => (BNF.TokenType.Backslash, "\\")
            | Token.Operator("=") => (BNF.TokenType.Eq, "=")
            | Token.Operator(":") => (BNF.TokenType.Colon, ":")
            | Token.Semicolon(_) => (BNF.TokenType.Semicolon, ";")
            | Token.StringLiteral(text) => (BNF.TokenType.String, text)
            | Token.WhiteSpace(text) => (BNF.TokenType.Whitespace, text)
            | _ =>
              Message.Error(token.Location, $"Unexpected token $token");
              (BNF.TokenType.Whitespace, "")
          };
        match (tokenType)
        {
          | BNF.TokenType.Whitespace
          | BNF.TokenType.Comment => ()
          | _ =>
            def bnfToken = BNF.Token(tokenType, loc, text);
            tokenList.Add(bnfToken);
        }
      }
    }
    
    public GetNextToken() : BNF.Token
    {
      if (index >= tokenList.Count - 2)
        BNF.Token(BNF.TokenType.Eof, TextLocator.Empty, "")
      else
      {
        def token = tokenList[index];
        ++index;
        token
      }
    }
    
    public EvalToken(token : BNF.Token, _ : IErrorLog) : BNF.Node
    {
      match (token.TokenType)
      {
        | BNF.TokenType.Identifier => BNF.Node.Id(token.Locator, token.Text)
        | BNF.TokenType.String => BNF.Node.String(token.Locator, token.Text)
        | _ => BNF.Node.Token(token.Locator)  
      }
    }
  }
}