﻿using System;
using System.IO;
using Nemerle.Utility;
using Rsdn.ParserTools;


namespace Rsdn.ParserTools.CSharp
{
  [assembly:Bnf(
  bnf
  {  
    lexer
    {
      Comment ::= SingleLineComment | DelimitedComment;
      subst SingleLineComment ::= "//" CommentInputChar* (NewLineChar | "\u000D" "\u000A");
      subst CommentInputChar ::= ~NewLineChar;
      subst NewLineChar ::= "\u000D" | "\u000A" | "\u0085" | "\u2028" | "\u2029" | eof;
      subst DelimitedComment ::= "/*" (~"*" | "*"+ ~"/")* "*"+ "/" | eof;
      Whitespace ::= (PureWhitespace | NewLineChar | eof)+;
      subst PureWhitespace ::= U_Zs | "\u0009" | "\u000B" | "\u000C";
      subst UnicodeEscapeSequence ::=
        "\\u" HexDigit HexDigit HexDigit HexDigit |
        "\\U" HexDigit HexDigit HexDigit HexDigit HexDigit HexDigit HexDigit HexDigit;
      subst HexDigit ::= "0".."9" | "A".."F" | "a".."f";
      Identifier ::= "@"? IdentifierStartCharacter IdentifierPartCharacter*;
      subst IdentifierStartCharacter ::= U_Ll | U_Lu | U_Lt | U_Lm | U_Lo | 
        UnicodeEscapeSequence | "_";
      subst IdentifierPartCharacter ::=
	      U_Nl | U_Ll | U_Lu | U_Lt | U_Lm | U_Lo | U_Nd | UnicodeEscapeSequence | U_Mc | U_Mn | "_";
      
      Keyword ::=
        "abstract" | "as" | "base" | "bool" | "break" | "byte" | "case" | "catch" |
        "char" | "checked" | "class" | "const" | "continue" | "decimal" | "default" |
        "delegate" | "do" | "double" | "else" | "enum" | "event" | "explicit" |
        "extern" | "false" | "finally" | "fixed" | "float" | "for" | "foreach" |
        "goto" | "if" | "implicit" | "in" | "int" | "interface" | "internal" | "is" |
        "lock" | "long" | "namespace" | "new" | "null" | "object" | "operator" |
        "out" | "override" | "params" | "private" | "protected" | "public" | "readonly" |
        "ref" | "return" | "sbyte" | "sealed" | "short" | "sizeof" | "stackalloc" |
        "static" | "string" | "struct" | "switch" | "this" | "throw" | "true" |
        "try" | "typeof" | "uint" | "ulong" | "unchecked" | "unsafe" | "ushort" |
        "using" | "virtual" | "void" | "volatile" | "while";
      
      RealLiteral ::=
        DecimalDigit+ "." DecimalDigit+ ExponentPart? RealTypeSuffix? |
        "." DecimalDigit+ ExponentPart? RealTypeSuffix? |
        DecimalDigit+ ExponentPart? RealTypeSuffix? |
        DecimalDigit+ RealTypeSuffix?;
      IntegerLiteral ::= DecIntLiteral | HexIntLiteral;
      subst DecIntLiteral ::= DecimalDigit+ IntegerTypeSuffix?;
      subst DecimalDigit ::= "0".."9";
      subst LSuff ::= "L" | "l";
      subst USuff ::= "U" | "u";
      subst IntegerTypeSuffix ::= USuff | LSuff | USuff LSuff | LSuff USuff;
      subst HexIntLiteral ::= ("0x" | "0X") HexDigit+ IntegerTypeSuffix?;
      subst ExponentPart ::= ("e" | "E") Sign? DecimalDigit+;
      subst Sign ::= "+" | "-";
      subst RealTypeSuffix ::= "F" | "f" | "D" | "d" | "M" | "m";
      CharLiteral ::= "'" Character "'";
      subst Character ::=
        SingleCharacter | SimpleEscapeSequence |
        HexEscapeSequence | UnicodeEscapeSequence;
      subst SingleCharacter ::= ~("\u0027" | "\u005C" | NewLineChar);
      subst SimpleEscapeSequence ::=
        "\\'" | "\\\"" | "\\\\" | "\\0" | "\\a" | "\\b" |
        "\\f" | "\\n" | "\\r" | "\\t" | "\\v";
      subst HexEscapeSequence ::= "\\x" HexDigit HexDigit? HexDigit? HexDigit?;
      StringLiteral ::= RegularStringLiteral | VerbatimStringLiteral;
      subst RegularStringLiteral ::= "\"" RegularStringChar* "\"";
      subst RegularStringChar ::= ~("\u0022" | "\u005C" | NewLineChar) |
        SimpleEscapeSequence | HexEscapeSequence | UnicodeEscapeSequence;
      subst VerbatimStringLiteral ::= "@\"" VerbatimStringChar* "\"";
      subst VerbatimStringChar ::= ~"\"" | "\"\"";
      
      Brace ::= "{" | "}";
      Paren ::= "[" | "]" | "(" | ")";
      Dot ::= ".";
      Comma ::= ",";
      Semicolon ::= ";";
      
      Symbol ::=
          ":" | "::" | "+" | "-" | "*" | "/" | "%" | "&" | "|" | "^" | "!" |
          "~" | "=" | "<" | ">" | "?" | "??" | "++" | "--" | "&&" | "||" |
          "<<" | ">>" | "==" | "!=" | "<=" | ">=" | "+=" | "-=" | "*=" |
          "/=" | "%=" | "&=" | "|=" | "^=" | "<<=" | ">>=" | "->";
      Preprocessor ::= "#" (any \ NewLineChar)*;

      /*
       *  Hacks
       */
      IntegerLiteral_Dot ::= DecIntLiteral ".";
    }
  }
  )]


  module MacroTest
  {
    Main() : void
    {
      def reader = StreamReader("test.txt");
      def lexer = LexelHelper(reader, "test.txt");
      def next_token()
      {
        def token = lexer.GetToken();
        match (token.TokenType)
        {
          | TokenType.Eof => ()
          | _ => Console.WriteLine($"$(token.TokenType) : $(token.Text)"); next_token()
        }
      }
      next_token();
    }
  }
}