using System;


namespace Rsdn.ParserTools.Test
{
  module Derivator
  {
    public Deriv(expr : Node, x : string) : Node
    {
      def mk_unary(op, a)
      {
        | ("-", Node.Function("-", p :: [])) => p
        | _ => Node.Function(TextLocator.Empty, op, [a])
      }
      def mk_bin(op, a, b)
      {
        | ("+", Node.Number(0.0f), p) => p
        | ("+", p, Node.Number(0.0f)) => p
        | ("-", Node.Number(0.0f), p) => mk_unary("-", p)
        | ("-", p, Node.Number(0.0f)) => p
        | ("*", _, Node.Number(0.0f) as q) => q
        | ("*", Node.Number(0.0f) as q, _) => q
        | ("*", p, Node.Number(1.0f)) => p
        | ("*", Node.Number(1.0f), p) => p
        | _ => Node.Function(TextLocator.Empty, op, [a, b])
      }
      match (expr)
      {
        | Node.Function("+", a :: b :: []) => mk_bin("+", Deriv(a, x), Deriv(b, x))
        | Node.Function("-", a :: b :: []) => mk_bin("-", Deriv(a, x), Deriv(b, x))
        | Node.Function("-", a :: []) => mk_unary("-", Deriv(a, x))
        | Node.Function("*", a :: b :: []) =>
          mk_bin("+", mk_bin("*", a, Deriv(b, x)), mk_bin("*", Deriv(a, x), b))
        | Node.Function("/", a :: b :: []) =>
          mk_bin("/",
            mk_bin("-", mk_bin("*", a, Deriv(b, x)), mk_bin("*", Deriv(a, x), b)),
            mk_bin("*", b, b))
        | Node.Id(id) when id == x => Node.Number(TextLocator.Empty, 1)
        | Node.Id(_) => Node.Number(TextLocator.Empty, 0)
        | Node.Number(_) => Node.Number(TextLocator.Empty, 0)
        | Node.Function("sin", a :: []) => mk_bin("*", Deriv(a, x), mk_unary("cos", a))
        | Node.Function("cos", a :: []) => mk_unary("-", mk_bin("*", Deriv(a, x), mk_unary("sin", a)))
        | _ => throw Exception()
      }
    }
  }
}