using Nemerle;
using System;
using Rsdn.ParserTools;


namespace Rsdn.ParserTools.Test
{
  public class NodeBuilder : INodeBuilder
  {
    log : IErrorLog;
    
    public this(log : IErrorLog)
    {
      this.log = log;
    }
    
    public Binary(loc : TextLocator, first : Node, op : Node, second : Node) : Node
    {
      match (op)
      {
        | Node.Token(TokenType.Plus) => Node.Function(loc, "+", [first, second])
        | Node.Token(TokenType.Minus) => Node.Function(loc, "-", [first, second])
        | Node.Token(TokenType.Star) => Node.Function(loc, "*", [first, second])
        | Node.Token(TokenType.Slash) => Node.Function(loc, "/", [first, second])
        | _ => throw Exception();
      }
    }
    
    public Prefix(loc : TextLocator, op : Node, expr : Node) : Node
    {
      match (op)
      {
        | Node.Token(TokenType.Minus) => Node.Function(loc, "-", [expr])
        | _ => throw Exception();
      }
    }
    
    public Function(loc : TextLocator, id : Node, args : Node) : Node
    {
      match ((id, args))
      {
        | (Node.Id(id), Node.List(args)) => Node.Function(loc, id, args)
        | _ => throw Exception();
      }
    }
    
    public List(loc : TextLocator, head : Node, tail : Node) : Node
    {
      match (tail)
      {
        | Node.List(l) => Node.List(loc, head :: l);
        | Node.Null() => Node.List(loc, [head]);
        | _ => throw Exception();
      }
    }
  }
}